/*
** Module   : ErrorWriterInteractive.java
** Abstract : Implementation for error writer for interactive mode.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20131119 Created initial version.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.util.*;

/**
 * Implements display outputs for interactive mode.
 */
public class ErrorWriterInteractive
implements ErrorWriter
{
   /** Client exports access. */
   private static ClientExports client = null; 
   
   private void getClientInstance()
   {
      if (client == null)
      {
         client = ThinClient.getInstance();
      }
   }
   
   public void displayError(String errmsg)
   {
      getClientInstance();
      
      if (client != null)
      {
         // we are running on the client
         ColorSpec cs = new ColorSpec(ColorSpec.COLOR_MESSAGES);
         
         client.message(errmsg, false, null, false, null, cs.convert());
      }
      else
      {
         // we are running on the server
         LogicalTerminal.message(errmsg);
      }
   }
   
   public void message(String errmsg)
   {
      LogicalTerminal.message(errmsg);
   }
   
   public void messageBox(String masterMsg, String title)
   {
      getClientInstance();

      if (title == null)
      {
         title = new String("Error");
      }
      
      if (client != null)
      {
         client.messageBox(new Object[] {masterMsg},
                           null,
                           LogicalTerminal.ALERT_ERROR,
                           LogicalTerminal.BTN_OK,
                           title,
                           (new ColorSpec(ColorSpec.COLOR_NORMAL)).convert()
                           );
      }
      else
      {
         LogicalTerminal.messageBox(masterMsg,
                                    LogicalTerminal.ALERT_ERROR,
                                    LogicalTerminal.BTN_OK,
                                    title);
      }
   }
   
   public void pause()
   {
      getClientInstance();
      
      if (client != null)
      {
         client.pause(-1, null, false);
      }
      else
      {
         LogicalTerminal.pause();
      }
      
      return;
   }
}
