/*
** Module   : BatchDriver.java
** Abstract : Batch mode based driver.
**
** Copyright (c) 2010-2011, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20131104 Created initial version
*/

package com.goldencode.p2j.ui.client.chui.driver.batch;

import java.util.*;

import com.goldencode.p2j.main.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.chui.driver.console.*;

/**
 *
 *
 */
public class BatchDriver
extends ConsoleDriver
implements ScreenDriver
{
   /** Reference to batch low-level driver. */
   private BatchHelper batchHelper = BatchHelper.getDefaultBatchHelper();

   /** Direct to screen output primitives. */
   private OutputPrimitives batchOut;

   /** Key processor. */
//   private KeyProcessor processor = new KeyProcessor();

   /** CHUI-specific widget factory. */
//   private WidgetFactory factory = new ChuiWidgetFactory();

   /**
    * Constructor. 
    */
   public BatchDriver()
   {
      super();
   }

   /**
    * Generate a beep signal. 
    * <p>
    * Implementation deoends on native layer and (in turn) from particular 
    * terminal. If terminal does not support this feature or sound in terminal
    * is turned off, there will be no audible feedback to user.
    */
   @Override
   public void beep()
   {
      batchHelper.outputOperation();
   }

   /**
    * Clear screen.
    */
   @Override
   public void clear()
   {
      batchHelper.outputOperation();
   }

   /**
    * Get instance of {@link OutputPrimitives} which provides direct access to 
    * required functionality of attached terminal.
    */
   @Override
   public OutputPrimitives getPrimitives()
   {
      return batchOut;
   }

   /**
    * Check if terminal supports colors. 
    * <p>
    * This implementation also checks if colors are enabled in low-level 
    * driver. If they are not (and usually this is so), then application 
    * will run in B/W mode even on color terminal.
    */
   @Override
   public boolean hasColors()
   {
      batchHelper.inputOperation();
      return false;
   }

   /**
    * Read the key from terminal and translate it into P2J internal 
    * representation. Translation details are in {@link KeyProcessor} class. 
    * Mouse events (which also returned from native layer) are filtered out 
    * because mouse is not supported.  
    */
/*   @Override
   public int readKey()
   {
//      helper.inputOperation();
      return ' ';
   }
*/
   /**
    * Reset screen mode.
    */
   @Override
   public void resetMode()
   {
      batchHelper.outputOperation();
   }

   /**
    * Resume terminal after {@link #suspend()} call.
    */
   @Override
   public void resume()
   {
      batchHelper.outputOperation();
   }

   /**
    * Suspend terminal and release it for use by other applications.
    */
   @Override
   public void suspend()
   {
      batchHelper.outputOperation();
   }

   /**
    * Initialize driver.
    */
   @Override
   public void init()
   {
      batchOut = new BatchPrimitives(ConsoleHelper.getDefaultConsoleHelper(), batchHelper);
   } 
}
